#ifndef _pc_foundation_reference_h_
#define _pc_foundation_reference_h_

/*
 *  Reference.h
 *  PcFoundation
 *
 *  Created by Patrick Coleman on 6/15/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

namespace PcFoundation {

/**
	smart pointer template.  works with objects that subclass ReferencedObject
	
	example code:
		Reference<T> r1 = new T();
		Reference<T> r2 = r1;
		r1 = new T;
		r2 = r1;
	where T is a subclass of ReferencedObject
*/
template <class T>
class Reference {

public:
	Reference(T* address = 0);
	~Reference();
	
	/**
		copy constructor
	*/
	Reference(const Reference<T>& other);
	
	/**
		assignment from another reference
	*/
	Reference& operator=(const Reference<T>& rhs);
	
	/**
		assignment from a pointer
	*/
	Reference& operator=(T* rhs);
	
	/**
		 dereferencing
	*/
	T* operator->();
	
	/**
		 dereferencing
	*/
	const T* operator->() const;
	
	/**
		bool conversion
	*/
	operator bool() const;
	
	/**
		comparison
	*/
	bool operator<(const Reference<T>& rhs) const;
	
private:
	T* address;
};


template <class T>
Reference<T>::Reference(T* address) {
	this->address = address;
	if (this->address) {
		(this->address->numReferences)++;
	} 
}

template <class T>
Reference<T>::Reference(const Reference<T>& other) {
	address = other.address;
	if (address) {
		(address->numReferences)++;
	} 
}

template <class T>
Reference<T>::~Reference() {
	if (address) {
		(address->numReferences)--;
		if (address->numReferences == 0) {
			delete address;
		}
	}
}

template <class T>
Reference<T>& Reference<T>::operator=(const Reference<T>& rhs) {
	// add count of assigneee
	if (rhs.address) {
		(rhs.address->numReferences)++;
	}
	
	// reduce count & maybe delete previous assignee
	if (address) {
		(address->numReferences)--;
		if (address->numReferences == 0) {
			delete address;
		}
	}
	
	// assign
	address = rhs.address;
	
	return *this;
}

template <class T>
Reference<T>& Reference<T>::operator=(T* rhs) {
	// add count of assigneee
	if (rhs) {
		(rhs->numReferences)++;
	}
	
	// reduce count & maybe delete previous assignee
	if (address) {
		(address->numReferences)--;
		if (address->numReferences == 0) {
			delete address;
		}
	}
	
	// assign
	address = rhs;
	
	return *this;
}

template <class T>
T* Reference<T>::operator->() {
	return address;
}

template <class T>
const T* Reference<T>::operator->() const {
	return address;
}

template <class T>
Reference<T>::operator bool() const {
	return (address != 0);
}
template <class T>
bool Reference<T>::operator<(const Reference<T>& rhs) const {
	return (address < rhs.address);
}

}

#endif