/*
 *  EulerRotation.cpp
 *  PcGeom
 *
 *  Created by Patrick Coleman on 6/11/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <math.h>

#include "EulerRotation.h"

namespace PcGeom {


EulerRotation::EulerRotation(double rx, double ry, double rz, RotationOrder order, bool local) {
	this->rx = rx;
	this->ry = ry;
	this->rz = rz;
	this->order = order;
	this->local = local;
}

EulerRotation::EulerRotation(const Matrix4& matrix, RotationOrder order, bool local) {
	this->order = order;
	this->local = local;

	// matrix factorization adapted from my python code
	rx = 0.0;
	ry = 0.0;
	rz = 0.0;
	
	if ((local and (order == XYZ)) or ((not local) and (order == ZYX))) { // MxMyMz factorization
		ry = sin(matrix[0][2]);
		if (ry < HALF_PI) {
			if (ry > - HALF_PI) {
				rx = atan2(-matrix[1][2], matrix[2][2]);
				rz = atan2(-matrix[0][1], matrix[0][0]);
			} else {
				rx = - atan2(matrix[1][0], matrix[1][1]);
				rz = 0.0;
			}
		} else {
			rx = atan2(matrix[1][0], matrix[1][1]);
			rz = 0.0;
		}
	} else if ((local and (order == XZY)) or ((not local) and (order == YZX))) { // MxMzMy factorization
		rz = asin(-matrix[0][1]);
		if (rz < HALF_PI) {
			if (rz > - HALF_PI) {
				rx = atan2(matrix[2][1], matrix[1][1]);
				ry = atan2(matrix[0][2], matrix[0][0]);
			} else {
				rx = atan2(-matrix[2][0], matrix[2][2]); // Eberly bug
				ry = 0.0;
			}
		} else {
			rx = atan2(matrix[2][0], matrix[2][2]);// # Eberly bug
			ry = 0.0;
		}
	} else if ((local and (order == YXZ)) or ((not local) and (order == ZXY))) { // MyMxMz factorization
		rx = asin(-matrix[1][2]);
		if (rx < HALF_PI) {
			if (rx > - HALF_PI) {
				ry = atan2(matrix[0][2], matrix[2][2]);
				rz = atan2(matrix[1][0], matrix[1][1]);
			} else {
				ry = atan2(- matrix[0][1], matrix[0][0]); // Eberly bug
				rz = 0.0;
			}
		} else {
			ry = - atan2(- matrix[0][1], matrix[0][0]); // Eberly Bug
			rz = 0.0;
		}
	} else if ((local and (order == YZX)) or ((not local) and (order == XZY))) { // MyMzMx factorization
		rz = asin(matrix[1][0]);
		if (rz < HALF_PI) {
			if (rz > - HALF_PI) {
				rx = atan2(- matrix[1][2], matrix[1][1]);
				ry = atan2(- matrix[2][0], matrix[0][0]);
			} else {
				rx = 0.0;
				ry = - atan2(matrix[2][1], matrix[2][2]);
			}
		} else {
			rx = 0.0;
			ry = atan2(matrix[2][1], matrix[2][2]);
		}
	} else if ((local and (order == ZXY)) or ((not local) and (order == YXZ))) { // MzMxMy factorization
		rx = asin(matrix[2][1]);
		if (rx < HALF_PI) {
			if (rx > - HALF_PI) {
				ry = atan2(- matrix[2][0], matrix[2][2]);
				rz = atan2(- matrix[0][1], matrix[1][1]);
			} else {
				ry = 0.0 ;
				rz = - atan2(matrix[0][2], matrix[0][0]);
			}
		} else {
			ry = 0.0;
			rz = atan2(matrix[0][2], matrix[0][0]);
		}
	} else { // MzMyMx factorization
		ry = asin(-matrix[2][0]);
		if (ry < HALF_PI) {
			if (ry > - HALF_PI) {
				rx = atan2(matrix[2][1], matrix[2][2]);
				rz = atan2(matrix[1][0], matrix[0][0]);
			} else {
				rx = 0.0;
				rz = - atan2(matrix[0][1], matrix[1][1]); // bug in eberly's code
			}
		} else {
			rx = 0.0; 
			rz = atan2(- matrix[0][1], matrix[0][2]);
		}
	}
}

Matrix4 EulerRotation::matrix() const {
	
	// x rotation matrix
	double cx = cos(rx);
	double sx = sin(rx);
	Matrix4	rxMat;
	rxMat[1][1] = cx;
	rxMat[2][2] = cx;
	rxMat[2][1] = sx;
	rxMat[1][2] = -sx;
		
	// y rotation matrix
	double cy = cos(ry);
	double sy = sin(ry);
	Matrix4 ryMat;
	ryMat[0][0] = cy;
	ryMat[0][2] = sy;
	ryMat[2][0] = -sy;
	ryMat[2][2] = cy;
	
	// z rotation matrix
	double cz = cos(rz);
	double sz = sin(rz);
	Matrix4 rzMat;
	rzMat[0][0] = cz;
	rzMat[0][1] = -sz;
	rzMat[1][0] = sz;
	rzMat[1][1] = cz;

	// concatenate appropriately
	if (local) {
		switch(order) {
			case PcGeom::XYZ:
				return rxMat * ryMat * rzMat;
			case PcGeom::XZY:
				return rxMat * rzMat * ryMat;
			case PcGeom::YXZ:
				return ryMat * rxMat * rzMat;
			case PcGeom::YZX:
				return ryMat * rzMat * rxMat;
			case PcGeom::ZXY:
				return rzMat * rxMat * ryMat;
			default: // ZYX
				return rzMat * ryMat * rxMat;
		}
	} else {
		switch(order) {
			case PcGeom::XYZ:
				return rzMat * ryMat * rxMat;
			case PcGeom::XZY:
				return ryMat * rzMat * rxMat;
			case PcGeom::YXZ:
				return rzMat * rxMat * ryMat;
			case PcGeom::YZX:
				return rxMat * rzMat * ryMat;
			case PcGeom::ZXY:
				return ryMat * rxMat * rzMat;
			default: // ZYX
				return rxMat * ryMat * rzMat;
		}
	
	}
}

Point3 EulerRotation::operator()(const Point3& p) const {
	return matrix() * p;
}

Vector3 EulerRotation::operator()(const Vector3& v) const {
	return matrix() * v;
}
	

}
