/*
 *  QRotation.cpp
 *  PcGeom
 *
 *  Created by Patrick Coleman on 6/11/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <math.h>
#include <iostream>
#include <PcMath/functions.h>
#include <PcGeom/functions.h>

#include "QRotation.h"

namespace PcGeom {

using std::cerr;
using PcMath::intMin;
using PcGeom::colocated;

QRotation::QRotation() {
	this->q= Quaternion(1.0, 0.0, 0.0, 0.0);
}

QRotation::QRotation(const Vector3& axis, double angle) {
	double halfAngle = 0.5 * angle;
	Vector3 qV = sin(halfAngle) * axis;
	q.a = cos(halfAngle);
	q.b = qV.x;
	q.c = qV.y;
	q.d = qV.z;
}

QRotation::QRotation(const Quaternion& q) {
	this->q = q;
}

Matrix4 QRotation::matrix() const {

	Matrix4 m;
	
	// adapted from my python code
	double x_2 = 2.0 * q.b;	// 2x
	double y_2 = 2.0 * q.c;	// 2y
	double z_2 = 2.0 * q.d;	// 2z
	double x2_2 = x_2 * q.b;	// 2x^2
	double y2_2 = y_2 * q.c;	// 2y^2
	double z2_2 = z_2 * q.d;	// 2z^2
	double xy_2 = x_2 * q.c;	// 2xy
	double xz_2 = x_2 * q.d;	//2xz
	double yz_2 = y_2 * q.d;	// 2yz
	double sx_2 = x_2 * q.a;	// 2sx
	double sy_2 = y_2 * q.a;	// 2sy
	double sz_2 = z_2 * q.a;	// 2sz
	
	m[0][0] = 1.0 - y2_2 -z2_2;
	m[0][1] = xy_2 - sz_2;
	m[0][2] = xz_2 + sy_2;
	m[1][0] = xy_2 + sz_2;
	m[1][1] = 1.0 - x2_2 - z2_2;
	m[1][2] = yz_2 - sx_2;
	m[2][0] = xz_2 - sy_2;
	m[2][1] = yz_2 + sx_2;
	m[2][2] = 1.0 - x2_2 - y2_2;

	return m;
}

Vector3 QRotation::operator()(const Vector3& v) const {
	Quaternion vQuat(0.0, v.x, v.y, v.z);
	Quaternion resQuat = q * vQuat * q.conjugate();
	return Vector3(resQuat.b, resQuat.c, resQuat.d);
}

Point3 QRotation::operator()(const Point3& p) const {
	Quaternion pQuat(0.0, p.x, p.y, p.z);
	Quaternion resQuat = q * pQuat * q.conjugate();
	return Point3(resQuat.b, resQuat.c, resQuat.d);
}
	
QRotation QRotation::operator+(const QRotation& rhs) const {

	return QRotation(q * rhs.q);
}

QRotation QRotation::operator-(const QRotation& rhs) const {
	return QRotation(q * rhs.q.conjugate());
}
	
EulerRotation QRotation::eulerRotation(RotationOrder order, bool local) const {
	return EulerRotation(matrix(), order, local);
}

QRotation QRotation::interpolate(const QRotation& rhs, double t, bool closest) const {
	
	double cosPhi = q.dot(rhs.q);
	Quaternion otherQ = rhs.q;
	
	if (closest and (cosPhi < 0.0)) {
		otherQ = - otherQ;
		cosPhi = q.dot(otherQ);
	}
	
	// numical inaccuracies can percolate this neglibably above 1.0 which causes acos to return nan
	if (cosPhi > 1.0) {
		cosPhi = 1.0;
	}
		
	double phi = acos(cosPhi);
		
	// test for interpolation against quats with zero angle
	if (phi <= 0.0) { // in [0, pi] by definition of acos
		return QRotation(q);
	}
				
	double invSinPhi = 1.0 / sin(phi);
	Quaternion unnormQ = (sin((1.0 - t) * phi) * invSinPhi * q + 
							 sin(t * phi) * invSinPhi * otherQ);
	return QRotation(unnormQ);
}

// ************************ static methods ******************************

QRotation QRotation::getAligningRotation(const Vector3& from, const Vector3& to, bool normalized) {

	double theta = from.angle(to);
	Vector3 axis = from ^ to;
	
	// no rotation if degenerate axis
	if (axis.lengthSquared() < 0.000001) {
		return QRotation();
	}
	
	if (not normalized) {
		axis = axis.normalize();
	}
	return QRotation(axis, theta);
}

QRotation QRotation::getAligningRotation(const Vector3& from1, const Vector3& from2,
										 const Vector3& to1, const Vector3& to2,
										 bool normalized) {
	
	// normalize if needed
	Vector3 from1N = from1;
	Vector3 from2N = from2;
	Vector3 to1N = to1;
	Vector3 to2N = to2;
	if (not normalized) {
		from1N = from1N.normalize();
		from2N = from2N.normalize();
		to1N = to1N.normalize();
		to2N = to2N.normalize();
	}
	
	// 1: find rotation to align the mean axes
	
	// get mean vecs
	double theta1 = from1N.angle(from2N);
	double theta2 = to1N.angle(to2N);
	Vector3 axis1 = (from1N ^ from2N).normalize();
	Vector3 axis2 = (to1N ^ to2N).normalize();
	QRotation qr1 = QRotation(axis1, 0.5 * theta1);
	QRotation qr2 = QRotation(axis2, 0.5 * theta2);
	Vector3 fromMeanVec = qr1(from1N);
	Vector3 toMeanVec = qr2(to1N);
	
	// compute rotaion that aligns the mean vectors 
	double thetaAlign = fromMeanVec.angle(toMeanVec);
	Vector3 axisAlign = (fromMeanVec ^ toMeanVec).normalize();
	QRotation qrAlign = QRotation(axisAlign, thetaAlign);
	
	// 2: find rotation to spin points about aligned mean vector
	
	// transporm inDirs to align them
	from1N = qrAlign(from1N);
	from2N = qrAlign(from2N);
	
	// find vectors connecting in pair and outpair
	Vector3 fromVec = from2N - from1N;
	Vector3 toVec = to2N - to1N;
	
	// find rotation that maps idVec to odVec
	Vector3 axisSpin = toMeanVec;
	double thetaSpin = fromVec.angle(toVec);
	Vector3 cp = (fromVec ^ toVec).normalize();
	if ((cp * axisSpin) < 0.0) {
		thetaSpin *= -1.0;	
	}
	QRotation qrSpin = QRotation(axisSpin, thetaSpin);
		
	// 3: concatenate the mean vector alignment and spin rotations
	return QRotation(qrSpin.q * qrAlign.q);
}


QRotation QRotation::getAligningRotation(const Vector3& from1, const Vector3& from2, const Vector3& from3,
										 const Vector3& to1, const Vector3& to2, const Vector3& to3,
										 bool normalized) {
	
	// normalize if needed
	Vector3 from1N = from1;
	Vector3 from2N = from2;
	Vector3 from3N = from3;
	Vector3 to1N = to1;
	Vector3 to2N = to2;
	Vector3 to3N = to3;
	if (not normalized) {
		from1N = from1N.normalize();
		from2N = from2N.normalize();
		from2N = from2N.normalize();
		to1N = to1N.normalize();
		to2N = to2N.normalize();
		to3N = to3N.normalize();
	}

	// 1: align normals of planes containing samples (this code will break down if they are colinear)
	
	// compute normals of planes
	Vector3 fromVec1 = from2N - from1N;
	Vector3 fromVec2 = from3N - from1N;
	Vector3 toVec1 = to2N - to1N;
	Vector3 toVec2 = to3N - to1N;
	Vector3 fromNormal = fromVec1 ^ fromVec2;
	Vector3 toNormal = toVec1 ^ toVec2;
	
	// compute aligning rotation
	Vector3 axisAlign = (fromNormal ^ toNormal).normalize();
	double thetaAlign = fromNormal.angle(toNormal);
	QRotation qrAlign = QRotation(axisAlign, thetaAlign);
	
	// 2: find best rotation to align planar vectors (they generaly lie in two parallel planes at this point)
	
	Vector3 axisSpin = toNormal.normalize();
	
	// aply aligning rotation to in dirs
	from1N = qrAlign(from1N);
	from2N = qrAlign(from2N);
	from3N = qrAlign(from3N);
	
	// project dirs to appropriate local plane
	from1N = from1N - (from1N * axisSpin) * axisSpin;
	from2N = from2N - (from2N * axisSpin) * axisSpin;
	from3N = from3N - (from3N * axisSpin) * axisSpin;
	to1N = to1N - (to1N * axisSpin) * axisSpin;
	to2N = to2N - (to2N * axisSpin) * axisSpin;
	to3N = to3N - (to3N * axisSpin) * axisSpin;
	
	// compute theta (closed form solution that minimizes angles)
	double num = (from1N ^ to1N) * axisSpin +
				 (from2N ^ to2N) * axisSpin +
				 (from3N ^ to3N) * axisSpin;	
	double denom = from1N * to1N + 
				   from2N * to2N + 
				   from3N * to3N;
	double thetaSpin = atan(num/denom);
	QRotation qrSpin(axisSpin, thetaSpin);
		
	// 3: concatenate the plane alignment alignment and spin rotations
	return QRotation(qrSpin.q * qrAlign.q); 
	
	
	/*
	ids = inDirections[:]
	ods = outDirections[:]
	if not normalized:
		ids = [id.normalize() for id in ids]
		ods = [od.normalize() for od in ods]

	# 1) align normals of planes containing samples (this code will break down if they are colinear)
	
	# and the cylons have a plane
	# compute normals of planes
	iVec1 = ids[1] - ids[0]
	iVec2 = ids[2] - ids[0]
	oVec1 = ods[1] - ods[0]
	oVec2 = ods[2] - ods[0]
	iN = iVec1 ^ iVec2
	oN = oVec1 ^ oVec2
	
	# reverse normals if needed to point toward majority of points (makes debugging visually easier)
	#iCentroid = 0.33333333 * (ids[0] + ids[1] + ids[2])
	#oCentroid = 0.33333333 * (ods[0] + ods[1] + ods[2])
	#if iN * iCentroid < 0.0:
	#	iN = -iN
	#if oN * oCentroid < 0.0:
	#	oN = -oN
	
	# compute aligning rotation
	axisAlign = (iN ^ oN).normalize()
	thetaAlign = iN.angle(oN)
	qrAlign = QRotation(axisAlign, thetaAlign)
	
	# 2) find best rotation to align planar vectors (they generaly lie in two parallel planes at this point)
	
	axisSpin = oN.normalize()
	
	# aply aligning rotation to in dirs
	ids = [qrAlign(v) for v in ids]
	
	# bring dirs to appropriate local plane
	ids = [v - (v * axisSpin) * axisSpin for v in ids]
	ods = [v - (v * axisSpin) * axisSpin for v in ods]
	
	# compute theta (closed form solution that minimizes angles)
	num = sum([(ids[i] ^ ods[i]) * axisSpin for i in range(3)])
	denom = sum([ids[i] * ods[i] for i in range(3)])
	thetaSpin = math.atan(num/denom)
	qrSpin = QRotation(axisSpin, thetaSpin)
		
	# 3) concatenate the plane alignment alignment and spin rotations
	return QRotation(q = qrSpin.q * qrAlign.q) #, iN, oN, axisAlign
	*/
}

QRotation QRotation::getAligningRotation(const vector<Vector3>& from, const vector<Vector3>& to, bool normalized) {

	int numVecs = intMin(from.size(), to.size());
	if (numVecs == 1 or colocated(from) or colocated(to)) {
		return getAligningRotation(from[0], to[0], normalized);
	} else if (numVecs == 2) {
		return getAligningRotation(from[0], from[1], to[0], to[1], normalized);
	} else if (numVecs == 3) {
		return getAligningRotation(from[0], from[1], from[2], to[0], to[1], to[2], normalized);
	} else {
		// TODO IMPLEMENT HORN'S FORMULA
		std::cerr << "WARNING: PcGeom::QRotation::getAligningRotation not implemented for general case\n";
	
		return getAligningRotation(from[0], from[1], from[2], to[0], to[1], to[2], normalized);
	}
}


// ************************ ops *****************************************


QRotation interpolate(const QRotation& lhs, const QRotation& rhs, double t) {
	return lhs.interpolate(rhs, t);
}

ostream& operator<<(ostream& output, const QRotation& qr) {
	output << qr.q;
    return output;
}

}
