/*
 *  functions.cpp
 *  PcGeom
 *
 *  Created by Patrick Coleman on 7/27/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include "functions.h"

namespace PcGeom {

bool colocated(const vector<Point3>& samples, double tol) {
	BoundingBox b(samples[0], samples[0]);
	for (int i = 0; i < samples.size(); i++) {
		b.addPoint(samples[i]);
	}
	return (b.diagonalSquared() <= tol * tol);
}

bool colocated(const vector<Vector3>& samples, double tol) {
	BoundingBox b(samples[0], samples[0]);
	for (int i = 0; i < samples.size(); i++) {
		b.addPoint(samples[i]);
	}
	return (b.diagonalSquared() <= tol * tol);
}

void getBoundingIndices(const vector<float>& knots, float t, int& priorIndex, int& nextIndex) {

	priorIndex = 0;
	nextIndex = 0;
		
	// t before knots or at start
	if (t <= knots[0]) {
		return;
	}
		
	// t after knots or at end
	int lastI = knots.size() - 1;
	if (t >= knots[lastI]) {
		priorIndex = lastI;
		nextIndex = lastI;
		return;
	}
		
	// test for t in knots
	for (int i = 1; i < knots.size() - 1; i++) {
		if (knots[i] == t) {
			priorIndex = i;
			nextIndex = i;
			return;
		}
	}
		
	// general check for prior index
	for (int i = 1; i < knots.size(); i++) {
		if (knots[i] > t) {
			priorIndex = i - 1;
			break;
		}
	}
		
	// set nextIndex, clamp to end
	nextIndex = priorIndex + 1;
	if (nextIndex == knots.size()) {
		nextIndex -= 1;
	}

}

float bezierEval(float v1, float v2, float v3, float v4, float t) {

	float t2 = t * t;
	float oneMinusT = 1.0 - t;
	float oneMinusT2 = oneMinusT * oneMinusT;
		
	float c0 = oneMinusT2 * oneMinusT;
	float c1 = 3.0 * t * oneMinusT2;
	float c2 = 3.0 * t2 * oneMinusT;
	float c3 = t2 * t;
		
	return (
		c0 * v1 + 
		c1 * v2 + 
		c2 * v3 + 
		c3 * v4
		);
}

Vector3 bezierEval(const Vector3& v1, const Vector3& v2, const Vector3& v3, 
				   const Vector3& v4, float t) {

	float t2 = t * t;
	float oneMinusT = 1.0 - t;
	float oneMinusT2 = oneMinusT * oneMinusT;
		
	float c0 = oneMinusT2 * oneMinusT;
	float c1 = 3.0 * t * oneMinusT2;
	float c2 = 3.0 * t2 * oneMinusT;
	float c3 = t2 * t;
		
	return (
		c0 * v1 + 
		c1 * v2 + 
		c2 * v3 + 
		c3 * v4
		);
}

Point3 bezierEval(const Point3& p1, const Point3& p2, const Point3& p3, 
				  const Point3& p4, float t) {
				  
	float t2 = t * t;
	float oneMinusT = 1.0 - t;
	float oneMinusT2 = oneMinusT * oneMinusT;
		
	float c0 = oneMinusT2 * oneMinusT;
	float c1 = 3.0 * t * oneMinusT2;
	float c2 = 3.0 * t2 * oneMinusT;
	float c3 = t2 * t;
		
	return (
		c0 * p1 + 
		c1 * p2 + 
		c2 * p3 + 
		c3 * p4
		);
}

QRotation bezierEval(const QRotation& q1, const QRotation& q2, const QRotation& q3,
					 const QRotation& q4, float t) {
	
	QRotation qOne1 = q1.interpolate(q2, t);
	QRotation qOne2 = q2.interpolate(q3, t);
	QRotation qOne3 = q3.interpolate(q4, t);
	
	QRotation qTwo1 = qOne1.interpolate(qOne2, t);
	QRotation qTwo2 = qOne2.interpolate(qOne3, t);
	
	return qTwo1.interpolate(qTwo2, t);
}



}


