/*
 *  ContactDetector.cpp
 *  PcMotion
 *
 *  Created by Patrick Coleman on 8/18/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <algorithm>

#include <PcSignal/functions.h>
#include <PcGeom/BoundingBox.h>

#include "ContactDetector.h"

namespace PcMotion {

using std::sort;
using PcSignal::getSpeedSquaredSignal;
using PcGeom::BoundingBox;

ContactDetector::ContactDetector(SampledMotion* motion, const RigidHierarchy& rh,
					const vector<string>& contactJoints,
					const vector<string>& mergeJoints,
					float tolerance, int combineWidth, float dt):
		motion(motion), rh(rh), contactJoints(contactJoints), mergeJoints(mergeJoints),
		contactsCached(false), mergedContactsCached(false),
		tolerance(tolerance), combineWidth(combineWidth), dt(dt) {
		
}

ContactMap ContactDetector::getContacts() {
	if (not contactsCached) {
		computeContacts();
	}
	
	return contacts;
}

ContactMap ContactDetector::getMergedContacts() {
	if (not mergedContactsCached) {
		computeMergedContacts();
	}
	
	return mergedContacts;
}

void ContactDetector::computeContacts() {
	contacts.clear();
	
	float tol2 = tolerance * tolerance;
	int ns = motion->numSamples();
	
	for (int i = 0; i < contactJoints.size(); i++) {
		
		if (not motion->hasTransAttr(contactJoints[i])) {
			continue;
		}
		
		// get speed signal
		vector<Point3> trajectory;
		for (int j = 0 ; j < ns; j++) {
			trajectory.push_back(motion->transSample(contactJoints[i], j));
		}
		vector<double> speedSig = getSpeedSquaredSignal(trajectory, dt);
	
		// get candidate indices using tolerance testing
		vector<int> candidateIndices;
		for (int j = 0; j < ns; j++) {
			if (speedSig[j] < tol2) {
				candidateIndices.push_back(j);
			}
		}

		if (candidateIndices.size() == 0) {
			continue;
		}

		// get contacts on the speed signal by searching and merging
		vector<Contact> jContacts;
		
		int startIndex = 0; 
		for (int j = 1; j < candidateIndices.size(); j++) {
		
			// break off a constraint and start a new one if index step > combineWidth
			if (candidateIndices[j] - candidateIndices[j - 1] > combineWidth) {
				int endIndex = j - 1;
			
				BoundingBox bbox(trajectory[candidateIndices[startIndex]],
								 trajectory[candidateIndices[startIndex]]);
				for (int k = candidateIndices[startIndex] + 1; k < candidateIndices[endIndex]; k++) {  
					bbox.addPoint(trajectory[k]);
				}
			
				jContacts.push_back(Contact(motion->startTime + candidateIndices[startIndex],
											motion->startTime + candidateIndices[endIndex],
											bbox.center()));
				startIndex = j;
			}
		}
		
		// append last constraint
		int endIndex = candidateIndices.size() - 1;
		
		BoundingBox bbox(trajectory[candidateIndices[startIndex]],
						 trajectory[candidateIndices[startIndex]]);
		for (int k = candidateIndices[startIndex] + 1; k < candidateIndices[endIndex]; k++) {  
			bbox.addPoint(trajectory[k]);
		}
				
		jContacts.push_back(Contact(motion->startTime + candidateIndices[startIndex],
							   motion->startTime + candidateIndices[endIndex],
							   bbox.center()));
	
		contacts[contactJoints[i]] = jContacts;
	}
	
	contactsCached = true;
}

void ContactDetector::computeMergedContacts() {
	if (not contactsCached) {
		computeContacts();
	}
	
	mergedContacts.clear();
	
	for (int mji = 0; mji < mergeJoints.size(); mji++) {
		
		if (not motion->hasTransAttr(mergeJoints[mji])) {
			continue;
		}
		
		//create a contact vector for all contacts on descendant joints
		vector<Contact> descContacts;
		ContactMap::iterator cmi;
		for (cmi = contacts.begin(); cmi != contacts.end(); cmi++) {
			if (rh.isDescendant(mergeJoints[mji], cmi->first)) {
				for (int ci = 0; ci < cmi->second.size(); ci++) {
					descContacts.push_back(cmi->second[ci]);
				}
			}
		}
		
		if (descContacts.size() == 0) {
			continue;
		}
		
		sort(descContacts.begin(), descContacts.end());
				
		// make contact vector of merged contacts
		vector<Contact> mergedContactsVec;
	
		// walk forward in time and break of contacts everytime a noncontact state is reached
		double startTime = descContacts[0].startTime;
		double endTime = descContacts[0].endTime;
		for (int dci = 1; dci < descContacts.size(); dci++) {
			if (descContacts[dci].startTime > endTime) {
				// break off a contact
				int startIndex = (int)(startTime - motion->startTime);
				int endIndex = (int)(endTime - motion->startTime);
				
				BoundingBox bbox(motion->transSample(mergeJoints[mji], startIndex),
								 motion->transSample(mergeJoints[mji], startIndex));
				for (int k = startIndex + 1; k <= endIndex; k++) {  
					bbox.addPoint(motion->transSample(mergeJoints[mji], k));
				}
			
				mergedContactsVec.push_back(Contact(startTime, endTime, bbox.center()));
				
				// update times
				startTime = descContacts[dci].startTime;
				endTime = descContacts[dci].endTime;
				
			} else if (descContacts[dci].endTime > endTime) {
				// push endTime back
				endTime = descContacts[dci].endTime;
			}
		}
		
		// break off last contact
		if (descContacts[descContacts.size() - 1].endTime > endTime) {
			endTime = descContacts[descContacts.size() - 1].endTime;
		}
		
		int startIndex = (int)(startTime - motion->startTime);
		int endIndex = (int)(endTime - motion->startTime);
				
		BoundingBox bbox(motion->transSample(mergeJoints[mji], startIndex),
						 motion->transSample(mergeJoints[mji], startIndex));
		for (int k = startIndex + 1; k <= endIndex; k++) {  
			bbox.addPoint(motion->transSample(mergeJoints[mji], k));
		}
			
		mergedContactsVec.push_back(Contact(startTime, endTime, bbox.center()));
	
		// add it to the merged contcats
		mergedContacts[mergeJoints[mji]] = mergedContactsVec;
	}
	
	mergedContactsCached = true;
}

}