/*
 *  ContactEnforcer.cpp
 *  PcMotion
 *
 *  Created by Patrick Coleman on 8/21/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <algorithm>

#include <PcMath/Vector3.h>
#include <PCGeom/PiecewiseSmoothCurve.h>

#include "ContactEnforcer.h"

namespace PcMotion {

using std::sort;
using PcMath::Vector3;
using PcGeom::PiecewiseSmoothFloatCurve;


// *************** ContactEnforcer public ops ****************

ContactEnforcer::ContactEnforcer(const ContactMap& contacts, const ContactMap& mergedContacts,
								 SampledMotion* referenceMotion, const RigidHierarchy& rh,
								 const vector<string>& ikStartJoints, const vector<string>& ikEndJoints) :
	contactMap(contacts), mergedContactMap(mergedContacts), referenceMotion(referenceMotion), rh(rh), 
	ikChainStartJoints(ikStartJoints), ikChainEndJoints(ikEndJoints),
	disableLocalCopy(false), disableVerticalCorrection(false) {
	
	verticalOffsetSignal.clear();

}

void ContactEnforcer::apply(SampledMotion* motion) const {
	
	// 1) copy feet and rotate to align in xz plane
	if (not disableLocalCopy) {
		copyIkChildrenLocal(motion);
	}
	
	// 2) apply up-down warp to keep feet on ground
	if (not disableVerticalCorrection) {
		adjustVerticalMotion(motion);
	}
	
	// 3) apply velocity warp to adjust for sliding
	
	// 4) stretch and ik for knee repositioning
	
	// 5) final foot cleanup
	//  consider allowing knot for maximum rotation difference during arc to be preserved
}

vector<float> ContactEnforcer::getVerticalOffsetSignal() const {
	return verticalOffsetSignal;
}

void ContactEnforcer::copyIkChildrenLocal(SampledMotion* motion) const {
	// copies local offsets of joints below ik end effector from reference motion to this motion 
	
	/*
	for each ankle
		for each descendant in rh
			for each Sample
				offset = desc at sample - ankle at sample from referenceMotion
				apply offset from ankle in motion
	*/
	
	// TODO change this to rotate points in xz plane
	
	int ns = motion->numSamples();
	int rns = referenceMotion->numSamples();
	if (rns < ns) {
		ns = rns;
	}
	
	for (int eji = 0; eji < ikChainEndJoints.size(); eji++) {
		vector<string> desc = rh.descendants(ikChainEndJoints[eji]);
		for (int di = 0; di < desc.size(); di++) {
			for (int si = 0; si < ns; si++) {
				Vector3 offset = referenceMotion->transSample(desc[di], si)
							   - referenceMotion->transSample(ikChainEndJoints[eji], si);
				motion->transSample(desc[di], si) = 
					motion->transSample(ikChainEndJoints[eji], si) + offset;
			}
		}
	}
	
}

void ContactEnforcer::adjustVerticalMotion(SampledMotion* motion) const {

	/*
		get vector<float> of sample times for each merged contact
		get vector<float> of sample offsets (vertical) for each merged contact
		merge them into one vector<float> each
		
		create a SmoothFloatCurve warp from these
		apply the warp to all translation channels of the motion
	*/

	// get sample times and sample offsets for each contact
	vector<vector<FloatSample> > sampleMap;
	getVerticalMergedContactSamples(motion, sampleMap);
	
	// merge indices and samples into one warp
	vector<FloatSample> samples;
	mergeFloatIndicesAndOffsets(motion, sampleMap, samples);

	// create a float curve warp
	vector<double> knots;
	vector<float> values;
	for (int i = 0; i < samples.size(); i++) {
		knots.push_back((double)samples[i].i);
		values.push_back(samples[i].v);
	}
	PiecewiseSmoothFloatCurve yWarp(knots, values);
	
	// apply the warp to the y comp of the entire motion
	int ns = motion->numSamples();

	for (int si = 0; si < ns; si++) {
		double u = (double)si;
		for (int i = 0; i < motion->numTransAttrs(); i++) {
			string tAttr = motion->transAttr(i);
			motion->transSample(tAttr, si) = motion->transSample(tAttr, si) + 
											Vector3(0.0, yWarp.value(u), 0.0);
		}
	}

	// cache the vertical offset signal
	verticalOffsetSignal.clear();
	for (int si = 0; si < ns; si++) {
		double u = (double)si;
		verticalOffsetSignal.push_back(yWarp.value(u));
	}

}

void ContactEnforcer::getVerticalMergedContactSamples(SampledMotion* motion, 
									 vector< vector<FloatSample> >& samples) const {
	samples.clear();
	
	int ns = motion->numSamples();
	
	ContactMap::iterator cmi;
	for (cmi = mergedContactMap.begin(); cmi != mergedContactMap.end(); cmi++) {
		for (int cmj = 0; cmj < cmi->second.size(); cmj++) {
			// process cmi->second[cmj] contact
		
			vector<FloatSample> contactSamples;
		
			int startIndex = (int)(cmi->second[cmj].startTime - motion->startTime);
			int endIndex = (int)(cmi->second[cmj].endTime - motion->startTime);
			for (int si = startIndex; si <= endIndex; si++) {
				if (si > 0 and si < ns) {
					float offset = referenceMotion->transSample(cmi->first, si).y - 
									motion->transSample(cmi->first, si).y;
					contactSamples.push_back(FloatSample(si, offset));
				}
			}
			
			if (contactSamples.size() > 0) {
				samples.push_back(contactSamples);
			}
		}
	}
	
}

void ContactEnforcer::mergeFloatIndicesAndOffsets(SampledMotion* motion,
												  const vector< vector<FloatSample> >& samples,
												  vector<FloatSample>& mergedSamples) const {
	mergedSamples.clear();
	
	if (samples.size() < 1) {
		return;
	}
	
	// merge all samples together and sort
	vector<FloatSample> allSamples;
	for (int ci = 0; ci < samples.size(); ci++) {
		for (int si = 0; si < samples[ci].size(); si++) {
			allSamples.push_back(samples[ci][si]);
		}
	}
	sort(allSamples.begin(), allSamples.end());
	
	// collapse recurring samples to maximum value
	int currentIndex = allSamples[0].i;
	float currentValue = allSamples[0].v;
	for (int si = 0; si < allSamples.size(); si++) {
		if (allSamples[si].i == currentIndex) {
			if (allSamples[si].v < currentValue) { // want most downward offset; i.e. most negative
				currentValue = allSamples[si].v;
			}
		} else {
			mergedSamples.push_back(FloatSample(currentIndex, currentValue));
			currentIndex = allSamples[si].i;
			currentValue = allSamples[si].v;
		}
	}
	
	// need last smaple
	mergedSamples.push_back(FloatSample(currentIndex, currentValue));
}


}