/*
 *  RigidHierarchy.cpp
 *  PcMotion
 *
 *  Created by Patrick Coleman on 6/24/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include "RigidHierarchy.h"

namespace PcMotion {

RHNode::RHNode(const string& name) {
	this->name = name;
	this->parent = 0;
}

RigidHierarchy::RigidHierarchy(const string& rootName) {
	rootNode = new RHNode(rootName);
	nodeMap[rootName] = rootNode;
}

string RigidHierarchy::root() const {
	return rootNode->name;
}

string RigidHierarchy::parent(const string& node) const {
	Reference<RHNode> parentNode = nodeMap[node]->parent;
	if (parentNode == 0) {
		return "";
	} else {
		return parentNode->name;
	}
}

string RigidHierarchy::child(const string& node, int i) const {
	return nodeMap[node]->children[i]->name;
}

int RigidHierarchy::numChildren(const string& node) const {
	return nodeMap[node]->children.size();
}

void RigidHierarchy::addChild(const string& node, const string& name) {
	if (nodeMap.find(node) != nodeMap.end()) {
		Reference<RHNode> n = new RHNode(name);
		n->parent = nodeMap[node];
		nodeMap[node]->children.push_back(n);
		nodeMap[name] = n;
	}
}

vector<string> RigidHierarchy::allNodes() const {
	// guarantees parents come before children

	vector<string> nodes;
	nodes.push_back(root());
	vector<string> desc = descendants(root());
	for (int i = 0; i < desc.size(); i++) {
		nodes.push_back(desc[i]);
	}
	return nodes;

	/* OLD CODE DOESN"T GUARANTEE PARENT FIRST 
	vector<string> nodes;
	map<string, Reference<RHNode> >::iterator mapIter = nodeMap.begin();
	for (; mapIter != nodeMap.end(); mapIter++) {
		nodes.push_back(mapIter->first);
	}
	return nodes;
	*/
}

vector<string> RigidHierarchy::descendants(const string& node) const {
	vector<string> desc;
	
	// immediate children
	for (int i = 0; i < numChildren(node); i++) {
		desc.push_back(child(node, i));
	}
	
	// recursive
	for (int i = 0; i < numChildren(node); i++) {
		vector<string> cDesc = descendants(child(node, i));
		for (int j = 0; j < cDesc.size(); j++) {
			desc.push_back(cDesc[j]);
		}
	}
	
	return desc;
}

bool RigidHierarchy::isDescendant(const string& node, const string& desc) {
	string p = parent(desc);
	while (p != "") {
		if (p == node) {
			return true;
		}
		p = parent(p);
	}
	return false;
}

RigidHierarchy* RigidHierarchy::copy() const {

	RigidHierarchy* newRh = new RigidHierarchy(rootNode->name);
	copyDescendants(rootNode->name, (*newRh));
	return newRh;
}	

void RigidHierarchy::copyDescendants(const string& node, RigidHierarchy& other) const {
	int numKids = numChildren(node);
	for (int i= 0; i < numKids; i++) {
		// copy child and call recursively
		other.addChild(node, child(node, i));
		copyDescendants(child(node, i), other);
	}
}

}

