/*
 *  SampledMotion.cpp
 *  PcMotion
 *
 *  Created by Patrick Coleman on 6/15/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include "SampledMotion.h"

namespace PcMotion {
	
// motion signal stuff
	
void SampledMotion::setTransSignal(const string& attribute, const TransSignal& signal) {
	transSignals[attribute] = signal;
}

void SampledMotion::setRotSignal(const string& attribute, const QRotSignal& signal) {
	rotSignals[attribute] = signal;
}

void SampledMotion::setTransSignals(const map<string, TransSignal>& signals) {
	transSignals.clear();
	transSignals = signals;
}

void SampledMotion::setRotSignals(const map<string, QRotSignal>& signals) {
	rotSignals.clear();
	rotSignals = signals;
}
		
// for sample access

Point3& SampledMotion::transSample(const string& attribute, int i) {
	return transSignals[attribute][i];
}

const Point3& SampledMotion::transSample(const string& attribute, int i) const {
	return transSignals[attribute][i];
}

QRotation& SampledMotion::rotSample(const string& attribute, int i) {
	return rotSignals[attribute][i];
}

const QRotation& SampledMotion::rotSample(const string& attribute, int i) const {
	return rotSignals[attribute][i];
}
	
// for time based evaluation, takes startTime and dTime into account

Point3 SampledMotion::transValue(const string& attribute, double t) const {
	return transSignals[attribute].value(t - startTime);
}

QRotation SampledMotion::rotValue(const string& attribute, double t) const {
	return rotSignals[attribute].value(t - startTime);
}

int SampledMotion::numSamples() const {
	if (numTransAttrs() == 0) {
		if (numRotAttrs() == 0) {
			return 0;
		} else {
			return rotSignals[transAttrs[0]].length();
		}
	} else {
		return transSignals[transAttrs[0]].length();
	}
}

double SampledMotion::endTime() const {
	return startTime + numSamples() * dTime;
}

Pose SampledMotion::getPoseAtIndex(int i) const {

	Pose p(transAttrs, rotAttrs);
	for (int j = 0; j < transAttrs.size(); j++) {
		p.setTransValue(transAttrs[j], transSignals[transAttrs[j]][i]);
	}
	for (int j = 0; j < rotAttrs.size(); j++) {
		p.setRotValue(rotAttrs[j], rotSignals[rotAttrs[j]][i]);
	}

	return p;
}

void SampledMotion::setPoseAtIndex(int i, const Pose& p) const {

	for (int j = 0; j < transAttrs.size(); j++) {
		transSignals[transAttrs[j]][i] = p.transValue(transAttrs[j]);
	}
	for (int j = 0; j < rotAttrs.size(); j++) {
		rotSignals[rotAttrs[j]][i] = p.rotValue(rotAttrs[j]);
	}

}

void SampledMotion::applyDisplacementMap(Motion* dispMap) {
	int ns = numSamples();

	for (int s = 0; s < ns; s++) {
		double t = (double)s + startTime;
		
		// apply trans attrs
		for (int i = 0; i < numTransAttrs(); i++) {
			string tAttr = transAttr(i);
			if (dispMap->hasTransAttr(tAttr)) {
				transSignals[tAttr][s] = transSignals[tAttr][s] + dispMap->transValue(tAttr, t);
			}
		}
		
		// apply rot attrs
		for (int i = 0; i < numRotAttrs(); i++) {
			string rAttr = rotAttr(i);
			if (dispMap->hasRotAttr(rAttr)) {
				rotSignals[rAttr][s] = dispMap->rotValue(rAttr, t) + rotSignals[rAttr][s];
			}
		}
	}
}

SampledMotion* SampledMotion::copy() const {
	SampledMotion* smCopy = new SampledMotion(startTime, transAttrs, rotAttrs, dTime);
	if (smCopy) {
		smCopy->setTransSignals(transSignals);
		smCopy->setRotSignals(rotSignals);
	}
	return smCopy;
}

}