/*
 *  SampledMotion.h
 *  PcMotion
 *
 *  Created by Patrick Coleman on 5/22/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#ifndef _pc_motion_sampled_motion_h_
#define _pc_motion_sampled_motion_h_

#include <string>
#include <map>
#include <PcGeom/Point3.h>
#include <PcGeom/QRotation.h>
#include <PcSignal/Signal.h>

#include "Motion.h"

namespace PcMotion {

using std::map;
using std::string;
using std::vector;
using PcGeom::Point3;
using PcGeom::QRotation;
using PcSignal::Signal;

typedef Signal<Point3, PcGeom::interpolate> TransSignal;
typedef Signal<QRotation, PcGeom::interpolate> QRotSignal;

/**
	uniformly sampled motion made up of motion signals
	startTime is the startTime of the motion
	dt is the time step, defaults to 1.0;
	dt doesn't currently work for values other than 1.0
	
	ALL SIGNALS SHOULD HAVE THE SAME NUMBER OF SAMPLES
*/
class SampledMotion: public Motion {

public:
	SampledMotion(double startTime, double dTime = 1.0f) : Motion(startTime) {
		this->dTime = dTime;
	}
	SampledMotion(double startTime, 
				  const vector<string>& transAttrs, 
				  const vector<string>& rotAttrs,
				  double dTime = 1.0f) : Motion(startTime, transAttrs, rotAttrs) {
		this->dTime = dTime;
	}
	
	
	// motion signal stuff
	
	void setTransSignal(const string& attribute, const TransSignal& signal);
	void setRotSignal(const string& attribute, const QRotSignal& signal);
	void setTransSignals(const map<string, TransSignal>& signals);
	void setRotSignals(const map<string, QRotSignal>& signals);
	
	
	// for sample access
	Point3& transSample(const string& attribute, int i);
	const Point3& transSample(const string& attribute, int i) const;
	QRotation& rotSample(const string& attribute, int i);
	const QRotation& rotSample(const string& attribute, int i) const;
	
	// for time based evaluation, takes startTime and dTime into account
	virtual Point3 transValue(const string& attribute, double t) const;
	virtual QRotation rotValue(const string& attribute, double t) const;
	
	int numSamples() const;
	double endTime() const;
	
	Pose getPoseAtIndex(int i) const;
	void setPoseAtIndex(int i, const Pose& p) const;
	
	/*
		applies the disp map motion to the samples of this motion
		rotations in the dispMap are subsequent to the current rotations
	*/
	void applyDisplacementMap(Motion* dispMap);
	
	SampledMotion* copy() const;
	
protected:
	double dTime;
	// mutable since keys accessors are not const, but operations follow const rules on semantic content of the maps
	mutable map<string, TransSignal> transSignals;  
	mutable map<string, QRotSignal> rotSignals;
};

}

#endif
