/*
 *  SkeletalPoseFitter.cpp
 *  PcMotion
 *
 *  Created by Patrick Coleman on 6/30/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <iostream>

#include <PcMath/Vector3.h>
#include <PcMath/Matrix4.h>

#include "SkeletalPoseFitter.h"
#include "RigidTransformHierarchy.h"

namespace PcMotion {

using PcMath::Vector3;
using PcMath::Matrix4;

SkeletalPoseFitter::SkeletalPoseFitter(const RigidHierarchy& hierarchy, const Pose& referencePose) :
	hierarchy(hierarchy), referencePose(referencePose) {
	
}

void SkeletalPoseFitter::setReferencePose(const Pose& rp) {
	referencePose = rp;
}

/*
void printPoint(const Point3& p) {
	std::cout << "[" << p.x << ", " << p.y << ", " << p.z << "]";
}*/

void printQr(const QRotation& qr) {
	std::cout << "[" << qr.q.a << ", " << qr.q.b << ", " << qr.q.c << ", " << qr.q.d << "]";
}

Pose SkeletalPoseFitter::operator()(const Pose& euclideanPose) const {

	vector<string> allNodes = hierarchy.allNodes();
	Pose newPose(referencePose);

	// scale translations to account for limb length changes
	for (int i = 0; i< allNodes.size(); i++) {
		string parent = hierarchy.parent(allNodes[i]);
		
		if (parent == "") { 
			// root; position it to match
			newPose.setTransValue(allNodes[i], euclideanPose.transValue(allNodes[i]));
		} else {
			// general limb; scale the limb length so that it can be 
			//rotated by its parent to matcch
			double desiredLength = euclideanPose.transValue(allNodes[i]).distance(
							euclideanPose.transValue(parent));
			Vector3 currentTranslate = newPose.transValue(allNodes[i]) - Point3();
			double currentLength = currentTranslate.length();
			currentTranslate *= (desiredLength / currentLength);
			newPose.setTransValue(allNodes[i], Point3(currentTranslate.x,
													  currentTranslate.y,
													  currentTranslate.z));
		}
	}
	
	// create the transform hierarchy for getting concatenated transforms
	RigidTransformHierarchy rth(hierarchy, newPose);
	
	// apply least squares rotation and residual in translation
	for (int i = 0; i< allNodes.size(); i++) {
		string node = allNodes[i];
		int numKids = hierarchy.numChildren(node);
		
		if (numKids > 0) {
			vector<Vector3> originalKidDirections;
			vector<Vector3> desiredKidDirections;
			
			// 1) get rotation to align kids from reference pose to the euclidean pose
			
			// collect local child vectors
			for (int k = 0; k < numKids; k++) {
				string kidNode = hierarchy.child(node, k);
				originalKidDirections.push_back(referencePose.transValue(kidNode) - Point3());
				Vector3 nkd = euclideanPose.transValue(kidNode) - euclideanPose.transValue(node);
				desiredKidDirections.push_back(nkd);
			}
			
			// bring euclidean child vectors into current local space of newPose
			MatrixTransform wit = rth.getInverseWorldTransform(node);
			for (int k = 0; k < numKids; k++) {
				desiredKidDirections[k] = wit(desiredKidDirections[k]);
			}
			
			// get rotation that aligns them
			QRotation alignQr = QRotation::getAligningRotation(originalKidDirections, desiredKidDirections);
			
			// append alignQr to current rotation
			QRotation currentQr = newPose.rotValue(node);
			QRotation newQr = currentQr + alignQr;// +  currentQr; seems backwords, maybe a problem with my theory?
			// no this is currect, because the inverse transform effectively removed the effect of current, so align HAS
			// to be applied first to align them in local space, and then current gets applied to restore the original motion!
			newPose.setRotValue(node, newQr);
			
			// update pose in rigid hierarchy
			rth.setPose(newPose);
			
			// 2) apply residual error from rotation fitting to child translations
			if (numKids > 1) {
			
				MatrixTransform wit = rth.getInverseWorldTransform(node);
				
				vector<Vector3> residuals;
				for (int k = 0; k < numKids; k++) {
					string kidNode = hierarchy.child(node, k);
					
					// get kid residual in world space
					MatrixTransform kwt = rth.getWorldTransform(kidNode);
					Matrix4 kwm = kwt.matrix();
					Point3 currentKidPos(kwm[0][3], kwm[1][3], kwm[2][3]);
					Point3 goalKidPos = euclideanPose.transValue(kidNode);
					Vector3 residual = goalKidPos - currentKidPos;
				
					// transform residuals to local space
					residual = wit(residual);
					
					// apply residual to child translation in newPose
					Point3 currentTranslate = newPose.transValue(kidNode);
					Point3 newTranslate = currentTranslate + residual;
					newPose.setTransValue(kidNode, newTranslate);
				}
				
				// update pose in rigid hierarchy
				rth.setPose(newPose);
			}
		} // else nothing if no kids
		
	
	}
	
	return newPose;
}


}