/*
 *  SpatialPoseSelector.cpp
 *  PcMotion
 *
 *  Created by Patrick Coleman on 7/26/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <PcSignal/functions.h>

#include "SpatialPoseSelector.h"

namespace PcMotion {

using PcSignal::semilocalMaxima;

SpatialPoseSelector::SpatialPoseSelector(const Matrix& dmat, 
										 float searchTol, int minSearchWidth,
										 int maximaWidth) :
	PoseSelector(NULL), dmat(dmat), searchTol(searchTol), minSearchWidth(minSearchWidth),
	maximaWidth(maximaWidth), extremenessComputed(false) {
	
}

vector<int> SpatialPoseSelector::getPoseIndices() {

	if (not extremenessComputed) {
		vector<int> nextBounds;
		vector<int> priorBounds;
		getBounds(priorBounds, nextBounds);
		
		computeExtremeness(priorBounds, nextBounds);
		
		extremenessComputed = true;
	}
	
	return semilocalMaxima(extremeness, maximaWidth);
}

vector<float> SpatialPoseSelector::getExtremeness() {
	if (not extremenessComputed) {
		vector<int> nextBounds;
		vector<int> priorBounds;
		getBounds(priorBounds, nextBounds);
		computeExtremeness(priorBounds, nextBounds);
		extremenessComputed = true;
	}
	return extremeness;
}

void SpatialPoseSelector::getBounds(vector<int>& priorBounds, vector<int>& nextBounds) {
	
	priorBounds.clear();
	nextBounds.clear();
		
	int sz = dmat.rowSize();
	float tolPercent = 1.0 - searchTol;
	
	for (int i = 0; i < sz; i++) {
		// get prior bound
		float maxPriorDist = dmat[i][i]; // zero
		int maxPriorIndex = i;
		bool priorSearchStopped = false;
		int j = i - 1;
		while ((j >= 0) and (not priorSearchStopped)) {
			
			// check for being further
			if (dmat[i][j] > maxPriorDist) {
				maxPriorDist = dmat[i][j];
				maxPriorIndex = j;
			}
			
			// if not further, check for stopping criterion
			else if ((i - j > minSearchWidth) and (maxPriorDist > 0.0)) {
				float percentBack = dmat[i][j] / maxPriorDist;
				if (percentBack < tolPercent) {
					priorSearchStopped = true;
				}
			}
			
			j -= 1;
		}
		
		priorBounds.push_back(maxPriorIndex);
		
		// get next bound
		float maxNextDist = dmat[i][i]; // zero
		int maxNextIndex = i;
		bool nextSearchStopped = false;
		
		j = i + 1;
		while ((j < sz) and (not nextSearchStopped)) {
			
			// check for being further
			if (dmat[i][j] > maxNextDist) {
				maxNextDist = dmat[i][j];
				maxNextIndex = j;
			}
			
			// if not further, check for stopping criterion
			else if ((j - i > minSearchWidth) and (maxNextDist > 0.0)) {
				float percentBack = dmat[i][j] / maxNextDist;
				if (percentBack < tolPercent) {
					nextSearchStopped = true;
				}
			}
		
			j += 1;
		}
		
		nextBounds.push_back(maxNextIndex);
	}
	
	/*
	bounds = []
	sz = distanceMatrix.rowSize()
	tolPercent = 1.0 - tol
	
	for i in range(sz):
		
		# get prior bound
		maxPriorDist = distanceMatrix[i][i] # zero
		maxPriorIndex = i
		priorSearchStopped = False
		
		j = i - 1
		while j >= 0 and not priorSearchStopped:
			
			# check for being further
			if distanceMatrix[i][j] > maxPriorDist:
				maxPriorDist = distanceMatrix[i][j]
				maxPriorIndex = j
			
			# if not further, check for stopping criterion
			elif i - j > minWidth and maxPriorDist > 0.0:
				percentBack = distanceMatrix[i][j] / maxPriorDist
				if percentBack < tolPercent:
					priorSearchStopped = True
		
			j -= 1
			
		# get next bound
		maxNextDist = distanceMatrix[i][i] # zero
		maxNextIndex = i
		nextSearchStopped = False
		
		j = i + 1
		while j < sz and not nextSearchStopped:
			
			# check for being further
			if distanceMatrix[i][j] > maxNextDist:
				maxNextDist = distanceMatrix[i][j]
				maxNextIndex = j
			
			# if not further, check for stopping criterion
			elif j - i > minWidth and maxNextDist > 0.0:
				percentBack = distanceMatrix[i][j] / maxNextDist
				if percentBack < tolPercent:
					nextSearchStopped = True
		
			j += 1
	
		# append bounds
		bounds.append((maxPriorIndex, maxNextIndex))
	
	return bounds
	*/
}

void SpatialPoseSelector::computeExtremeness(vector<int>& priorBounds, vector<int>& nextBounds) {
	int sz = dmat.rowSize();
	extremeness.clear();
	
	for (int i = 0; i < sz; i++) {
		extremeness.push_back(dmat[i][priorBounds[i]] + dmat[i][nextBounds[i]]);
	}
	
	/*
	for i in range(sz):
		# integral version
		#val = 0.0
		#for j in range(bounds[i][0], bounds[i][1] + 1):
		#	val += distanceMatrix[i][j]
		#extremeness.append(val)#/ float(bounds[i][1] - bounds[i][0] + 1))
		
		# just the distant points version
		extremeness.append(distanceMatrix[i][bounds[i][0]] + distanceMatrix[i][bounds[i][1]])
	*/
}

}