/*
 *  SpatialPoseExaggerator.cpp
 *  PcPoseExaggeration
 *
 *  Created by Patrick Coleman on 7/27/09.
 *  Copyright 2009 patrick coleman. All rights reserved.
 *
 */

#include <string>
#include <iostream>

#include <PcMath/Vector3.h>
#include <PcGeom/Point3.h>
#include <PcMotion/CurveMotion.h>

#include "SpatialPoseExaggerator.h"

namespace PcPoseExaggeration {

using std::string;
using PcMath::Vector3;
using PcGeom::Point3;
using PcMotion::CurveMotion;

SpatialPoseExaggerator::SpatialPoseExaggerator(const RigidHierarchy& rh,
											   const vector<int>& extremePoseIndices,
											   float exaggeration, float bias,
											   bool rescalePoses, bool rescaleMotion
											   ) :
	rh(rh), extremePoseIndices(extremePoseIndices),
	exaggeration(exaggeration), bias(bias),
	rescalePoses(rescalePoses), rescaleMotion(rescaleMotion) {
	
}

void printPose(const Pose& p) {
	//std::cout << "\t"; 
	for (int i = 0; i < p.numTransAttrs(); i++) {
		Point3 v = p.transValue(p.transAttr(i));
		std::cout << "\t\t" <<p.transAttr(i) << ": [" << v.x << ", " << v.y << ", " << v.z << "]\n";
	}
	std::cout << "\n";
}

void SpatialPoseExaggerator::apply(SampledMotion* originalMotion) const {
	
	int numPoses = extremePoseIndices.size();
	
	// get original extreme poses
	originalPoses.clear();
	for (int i = 0; i < numPoses; i++) {
		originalPoses.push_back(originalMotion->getPoseAtIndex(extremePoseIndices[i]));
	}
	
	// exaggerate extreme pose
	exagPoses = exaggeratePoses(originalPoses);
	
	// scale cartesian poses
	if (rescalePoses) {
		for (int i = 0; i < numPoses; i++) {
			correctPoseLengths(exagPoses[i], originalPoses[i]);
		}
	}
	
	// get displacement map values as difference poses
	differencePoses.clear();
	for (int i = 0; i < exagPoses.size(); i++) {
		differencePoses.push_back(exagPoses[i] - originalPoses[i]);
	}
	
	// apply as a trajectory displacement map to originalMotion
	vector<double> times;
	for (int i = 0; i < extremePoseIndices.size(); i++) {
		times.push_back(extremePoseIndices[i] + originalMotion->startTime);
	}
	
	CurveMotion* dmap = new CurveMotion(originalMotion->startTime, times, differencePoses);  
	SampledMotion* motionCopy;
	if (rescaleMotion) {
		motionCopy = originalMotion->copy();
	}
	originalMotion->applyDisplacementMap(dmap);
	delete dmap;
		
	// correct for limb lengths after dmapping
	if (rescaleMotion) {
		correctMotionLengths(originalMotion, motionCopy);
		delete motionCopy;
	}
}

vector<Pose> SpatialPoseExaggerator::getExaggeratedPoses() const {

	return exagPoses;
}

vector<Pose> SpatialPoseExaggerator::getOriginalPoses() const {

	return originalPoses;
}

vector<Pose> SpatialPoseExaggerator::getDifferencePoses() const {

	return differencePoses;
}

Pose SpatialPoseExaggerator::subtractRoot(const Pose& p) const {
	
	Pose retPose(p);
	string root = rh.root();
	Point3 rootPos = p.transValue(root);
	retPose.setTransValue(root, rootPos);
	
	for (int i = 0; i < p.numTransAttrs(); i++) {
		string tAttr = p.transAttr(i);
		if (tAttr != root) {
			Point3 pos = p.transValue(tAttr);
			Vector3 localPos = pos - rootPos;
			retPose.setTransValue(tAttr, Point3(localPos.x, localPos.y, localPos.z));		
		}
	}
	
	return retPose;
}

Pose SpatialPoseExaggerator::addRoot(const Pose& p) const {
	Pose retPose(p);
	
	string root = rh.root();
	Point3 rootPos = p.transValue(root);
	retPose.setTransValue(root, rootPos);
	
	for (int i = 0; i < p.numTransAttrs(); i++) {
		string tAttr = p.transAttr(i);
		if (tAttr != root) {
			Point3 pos = rootPos + p.transValue(tAttr);
			retPose.setTransValue(tAttr, pos);		
		}
	}
	
	return retPose;

}



vector<Pose> SpatialPoseExaggerator::exaggeratePoses(const vector<Pose>& originalPoses) const {
	
	float e = 0.5 * (exaggeration - 1.0);
	// subtract root
	vector<Pose> localPoses;
	for (int i = 0; i< originalPoses.size(); i++) {
		localPoses.push_back(subtractRoot(originalPoses[i]));
	}

	

	// exag
	vector<Pose> exagPoses;
	for (int i = 0; i< localPoses.size(); i++) {
		Pose ePose(localPoses[i]);
		string root = rh.root();
		ePose.setTransValue(root, localPoses[i].transValue(root));
		for (int j = 0; j < localPoses[i].numTransAttrs(); j++) {
			string tAttr = localPoses[i].transAttr(j);
			
			if (tAttr != root) {
				if (i == 0) {
					Point3 tVal = localPoses[i].transValue(tAttr);
					Point3 ntVal = localPoses[i + 1].transValue(tAttr);
			
					Point3 newPos = tVal + e * (tVal - ntVal);
				
					ePose.setTransValue(tAttr, newPos);	
				} else if (i == localPoses.size() - 1) {
					Point3 tVal = localPoses[i].transValue(tAttr);
					Point3 ptVal = localPoses[i - 1].transValue(tAttr);
			
					Point3 newPos = tVal + e * (tVal - ptVal);
				
					ePose.setTransValue(tAttr, newPos);	
				} else {
					Point3 tVal = localPoses[i].transValue(tAttr);
					Point3 ntVal = localPoses[i + 1].transValue(tAttr);
					Point3 ptVal = localPoses[i - 1].transValue(tAttr);
			
					Point3 newPos = tVal + 
									bias * e * (tVal - ntVal) + 
									(1.0 - bias) * e * (tVal - ptVal);

					ePose.setTransValue(tAttr, newPos);	
				}	
			}
		}
		exagPoses.push_back(ePose);
	}
	
	// add root back
	for (int i = 0; i< exagPoses.size(); i++) {
		exagPoses[i] = addRoot(exagPoses[i]);
	}
	
	return exagPoses;
}

map<const string, Vector3> SpatialPoseExaggerator::getOffsets(const Pose& p) const {

	map<const string, Vector3> offsets;

	string root = rh.root();	
	vector<string> nodes = rh.allNodes();
	
	for (int i = 0; i < nodes.size(); i++) {
		if (nodes[i] != root) {
			Point3 pNode = p.transValue(nodes[i]);
			Point3 pParent = p.transValue(rh.parent(nodes[i]));
			offsets[nodes[i]] = pNode - pParent; 
		}
	}
	
	return offsets;
}

void SpatialPoseExaggerator::correctOffsets(map<const string, Vector3>& offsets, 
											const Pose& referencePose) const {
	
	for (int i = 0; i < referencePose.numTransAttrs(); i++) {
		string node = referencePose.transAttr(i);
		if (node != rh.root()) {
			string parent = rh.parent(node);
			Point3 pos = referencePose.transValue(node);
			Point3 parentPos = referencePose.transValue(parent);
			double desiredLen = pos.distance(parentPos);
			double currentLen = offsets[node].length();
			double scale = 1.0;
			if (currentLen > 0.000001) {
				scale = desiredLen / currentLen;
			}
			offsets[node] = scale * offsets[node];
		}
	}
}

void SpatialPoseExaggerator::setOffsetsToPose(map<const string, Vector3>& offsets, 
											  Pose& pose,
											  const string& belowThisNode) const {
	for (int i = 0; i < rh.numChildren(belowThisNode); i++) {
		string child = rh.child(belowThisNode, i);
		Point3 newPos = pose.transValue(belowThisNode) + offsets[child];
		pose.setTransValue(child, newPos); 
		setOffsetsToPose(offsets, pose, child);
	}
}

void SpatialPoseExaggerator::correctPoseLengths(Pose& pose, 
											   const Pose& referencePose) const {
	// convert pose to string:vector map
	map<const string, Vector3> offsets = getOffsets(pose);
	
	// correct lengths of vectors wrt reference pose
	correctOffsets(offsets, referencePose);
	
	// set new values from string:vector map
	setOffsetsToPose(offsets, pose, rh.root());
}	
	
void SpatialPoseExaggerator::correctMotionLengths(SampledMotion* motion, 
											const SampledMotion* referenceMotion) const {

	for (int i = 0; i < motion->numSamples(); i++) {
		Pose p = motion->getPoseAtIndex(i);
		Pose rp = referenceMotion->getPoseAtIndex(i);
		correctPoseLengths(p, rp);
		motion->setPoseAtIndex(i, p);
	}
}

}